from inspect import iscoroutinefunction, isfunction
from nurse.exceptions import DependencyError
from .service_catalog import ServiceCatalog
from typing import List


def serve(user_class, through=None) -> None:
    """
    Add an instance of a user-defined class to Nurse's services catalog.
    By default, a dependency is registered for its concrete type, but an interface can be provided.

    :param user_class: User-defined class instance
    :param through: An interface used to access the user class
                   (must be a direct or indirect parent class)
    """

    through = through or user_class.__class__

    if not issubclass(user_class.__class__, through):
        raise ValueError(f"Class {user_class} must be a subclass of {through}.")

    ServiceCatalog.get_instance()._services[through] = user_class


def clear() -> None:
    """
    Remove all existing registered services
    """
    ServiceCatalog.get_instance().clear()


def get(service_type):
    """
    Retrieve a service from the service catalog.

    :Example:

    nurse.serve(SSHClient())

    ssh_client = nurse.get(SSHClient)
    """
    return ServiceCatalog.get_instance()._services.get(service_type)


def inject(*items_to_inject: List[str]):
    """
    A decorator that injects dependencies into every instances of a user-defined class or method

    :Example:

    class SSHClient:

        def connect(self):
            pass

    @nurse.inject("ssh_client")
    class Server:
        ssh_client: SSHClient

    server = Server()
    server.ssh_client.connect()


    @nurse.inject("client")
    def send(client: SSHClient):
        client.send("Hello World !")

    nurse.serve(SSHClient())
    send()


    @nurse.inject("client")
    async def send(client: SSHClient):
        await client.send("Hello World !")

    nurse.serve(SSHClient())
    asyncio.run(send())
    """

    def decorator(decorated):
        service_catalog = ServiceCatalog.get_instance()
        if isinstance(decorated, type):
            return inject_class(decorated, service_catalog, items_to_inject)
        elif iscoroutinefunction(decorated):
            return inject_async_function(decorated, service_catalog, items_to_inject)
        elif isfunction(decorated):
            return inject_function(decorated, service_catalog, items_to_inject)

        raise NotImplementedError("user-defined class or function can't be injected.")

    return decorator


def inject_class(decorated_class, service_catalog, field_to_inject):
    def init_decorator(self, *args, **kwargs):
        for param_to_inject in field_to_inject:
            service = get_service(service_catalog, decorated_class, param_to_inject)
            setattr(self, param_to_inject, service)

        return init_decorator.decorated_init(self, *args, **kwargs)

    init_decorator.decorated_init = decorated_class.__init__
    decorated_class.__init__ = init_decorator

    return decorated_class


def inject_function(decorated_func, service_catalog, args_to_inject):
    def decorator(*args, **kwargs):
        injected_args = {}
        for param_to_inject in args_to_inject:
            service = get_service(service_catalog, decorated_func, param_to_inject)
            injected_args.setdefault(param_to_inject, service)

        return decorated_func(*args, **injected_args, **kwargs)

    return decorator


def inject_async_function(decorated_func, service_catalog, args_to_inject):
    async def decorator(*args, **kwargs):
        injected_args = {}
        for param_to_inject in args_to_inject:
            service = get_service(service_catalog, decorated_func, param_to_inject)
            injected_args.setdefault(param_to_inject, service)

        return await decorated_func(*args, **injected_args, **kwargs)

    return decorator


def get_service(service_catalog: ServiceCatalog, decorated_obj, param_to_inject: str):
    service_type = decorated_obj.__annotations__.get(param_to_inject)
    if not service_type:
        raise DependencyError(f"Args `{param_to_inject}` must be typed to be injected.")

    service = service_catalog._services.get(service_type)
    if not service:
        raise DependencyError(
            f"Dependency `{service_type}` for `{param_to_inject}` was not found."
        )

    return service



class ServiceCatalog:
    """
    A singleton service catalog.
    """

    __slots__ = ("_services",)
    instance = None

    def __init__(self) -> None:
        self._services = {}

    def clear(self) -> None:
        self._services.clear()

    @classmethod
    def get_instance(cls) -> "ServiceCatalog":
        if cls.instance is None:
            cls.instance = cls()

        return cls.instance



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nurse']

package_data = \
{'': ['*']}

install_requires = \
['flake8>=3.7.9,<4.0.0']

setup_kwargs = {
    'name': 'nurse',
    'version': '0.5.0',
    'description': 'A thoughtful dependency injection framework 💉',
    'long_description': 'Nurse\n=====\n\n.. image:: https://img.shields.io/badge/license-public%20domain-ff69b4.svg\n    :target: https://github.com/ZeroGachis/nurse#license\n\n\n.. image:: https://img.shields.io/badge/pypi-v0.3.1-blue.svg\n    :target: https://pypi.org/project/nurse/\n\n\nOutline\n~~~~~~~\n\n1. `Overview <https://github.com/ZeroGachis/nurse#overview>`_\n2. `Installation <https://github.com/ZeroGachis/nurse#installation>`_\n3. `Usage <https://github.com/ZeroGachis/nurse#usage>`_\n4. `License <https://github.com/ZeroGachis/nurse#license>`_\n\n\nOverview\n~~~~~~~~\n\n\n**Nurse** is a **dependency injection framework** with a small API that uses\ntype annotations to manage dependencies in your codebase.\n\n\nInstallation\n~~~~~~~~~~~~\n\n**Nurse** is a Python3-only module that you can install via `Poetry <https://github.com/sdispater/poetry>`_\n\n.. code:: sh\n\n    poetry add nurse\n\n\nIt can also be installed with `pip`\n\n.. code:: sh\n\n    pip3 install nurse\n\n\nUsage\n~~~~~\n\n**Nurse** stores the available dependencies into a service catalog, that needs to be\nfilled-in generally at the startup of your application.\n\n.. code:: python3\n\n    import nurse\n    \n    # A user defined class that will be used accross your application\n    class Player:\n        \n        @property\n        def name(self) -> str:\n            return "Leeroy Jenkins"\n\n    # Now, add it to nurse service catalog in order to use it later in your application\n    nurse.serve(Player())\n\nBy default, dependencies are referenced by their concrete type, but you can also serve them\nvia one of their parent class.\n\n.. code:: python3\n\n    import nurse\n\n    class Animal:\n        pass\n\n    class AngryAnimal(Animal):\n\n        @property\n        def roar(self) -> str:\n            return "Grrr! 🦁"\n\n    nurse.serve(AngryAnimal(), through=Animal)\n\nOnce you filled-in the service catalog with your different components, your can declare them as dependencies\nto any of your class.\n\n.. code:: python3\n\n    @nurse.inject("player")\n    class Game:\n        player: Player\n        enemy: Animal\n\n        def welcome_hero(self):\n            print(f"Welcome {self.player.name} !")\n    \n        def summon_monster(self):\n            print(self.enemy.roar)\n\n    Game = Game()\n    game.welcome_hero()\n    # Welcome Leeroy Jenkins !\n    game.summon_monster()\n    # Grrr! 🦁\n\n\nOr in any function\n\n.. code:: python3\n\n    @nurse.inject(\'enemy\')\n    def summon_monster(enemy: Animal):\n        print(enemy.roar)\n\n    summon_monster()\n    # Grrr! 🦁\n\n\nAnd it works with async function as well !\n\n.. code:: python3\n\n    import asyncio\n\n    @nurse.inject(\'enemy\')\n    async def summon_monster(enemy: Animal):\n        print(enemy.roar)\n\n    asyncio.run(summon_monster())\n    # Grrr! 🦁\n\n\nFinally, you can also retrieve a service without using a decorator\n\n.. code:: python3\n\n    enemy = nurse.get(Animal)\n    print(enemy.roar)\n    # Grrr! 🦁\n\n\nLicense\n~~~~~~~\n\n**Nurse** is released into the Public Domain. 🎉\n',
    'author': 'ducdetronquito',
    'author_email': 'g.paulet@zero-gachis.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ZeroGachis/nurse',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
